"""Utility functions for the account_keeping app."""
from django.utils.timezone import datetime, now

from six import string_types


def get_date(value):
    """
    Returns the given field as a DateTime object.

    This is necessary because Postgres and SQLite return different values
    for datetime columns (DateTime vs. string).

    """
    if isinstance(value, string_types):
        return datetime.strptime(value, '%Y-%m-%d')
    return value


def get_months_of_year(year):
    """
    Returns the number of months that have already passed in the given year.

    This is useful for calculating averages on the year view. For past years,
    we should divide by 12, but for the current year, we should divide by
    the current month.

    """
    current_year = now().year
    if year == current_year:
        return now().month
    if year > current_year:
        return 1
    if year < current_year:
        return 12



"""Tests for the management commands of the ``account_keeping`` app."""
from datetime import date
from os import path

from django.core.management import call_command, CommandError
from django.test import TestCase

from mixer.backend.django import mixer


class CommandTestCase(TestCase):
    def test_collect_invoices(self):
        transaction = mixer.blend('account_keeping.Transaction',
                                  transaction_date=date.today())
        mixer.blend('account_keeping.Transaction', parent=transaction)
        mixer.blend('account_keeping.Transaction',
                    transaction_date=date.today(), account=transaction.account)
        call_command('collect_invoices', account=transaction.account.slug,
                     start_date=date.today().strftime('%Y-%m-%d'),
                     end_date=date.today().strftime('%Y-%m-%d'),
                     output='./foo')

    def test_importer_mmex(self):
        currency = mixer.blend('currency_history.Currency')
        call_command('importer_mmex',
                     account=mixer.blend('account_keeping.Account').slug,
                     currency=currency.iso_code,
                     filepath=path.abspath(path.join(path.dirname(
                         path.dirname(__file__)), 'tests', 'test_file.csv')))
        with self.assertRaises(CommandError):
            call_command('importer_mmex', currency='FOO')



"""
These settings are used by the ``manage.py`` command.
With normal tests we want to use the fastest possible way which is an
in-memory sqlite database but if you want to create South migrations you
need a persistant database.
Unfortunately there seems to be an issue with either South or syncdb so that
defining two routers ("default" and "south") does not work.
"""
from distutils.version import StrictVersion

import django

from .test_settings import *  # NOQA


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'db.sqlite',
    }
}

django_version = django.get_version()
if StrictVersion(django_version) < StrictVersion('1.7'):
    INSTALLED_APPS.append('south', )  # NOQA



"""Views for the account_keeping app."""
import decimal
from datetime import date

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse, reverse_lazy
from django.db import connection
from django.db.models import Q, Sum
from django.http import HttpResponse
from django.shortcuts import redirect
from django.template.defaultfilters import date as date_filter
from django.utils.decorators import method_decorator
from django.utils.timezone import now
from django.views import generic

from currency_history.models import Currency, CurrencyRateHistory
from dateutil import relativedelta

from . import forms
from . import models
from . import utils
from .freckle_api import get_unpaid_invoices_with_transactions
from .utils import get_date as d


DEPOSIT = models.Transaction.TRANSACTION_TYPES['deposit']
WITHDRAWAL = models.Transaction.TRANSACTION_TYPES['withdrawal']


class LoginRequiredMixin(object):
    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        return super(LoginRequiredMixin, self).dispatch(
            request, *args, **kwargs)


class AccountsViewMixin(object):
    """
    Mixin that collects all data to show all transactions for all accounts.

    """
    template_name = 'account_keeping/accounts_view.html'

    def get_context_data(self, **kwargs):
        ctx = super(AccountsViewMixin, self).get_context_data(**kwargs)
        accounts = models.Account.objects.all()
        account_transactions = []
        totals = {
            'amount_net': 0,
            'amount_gross': 0,
            'expenses_net': 0,
            'expenses_gross': 0,
            'income_net': 0,
            'income_gross': 0,
        }
        base_currency = getattr(settings, 'BASE_CURRENCY', 'EUR')
        for account in accounts:
            rate = 1
            if not account.currency.iso_code == base_currency:
                rate = self.get_rate(account.currency)

            account_balance = account.get_balance(self.month)

            qs = self.get_transactions(account)

            amount_net_sum = qs.aggregate(
                Sum('value_net'))['value_net__sum'] or 0
            amount_gross_sum = qs.aggregate(
                Sum('value_gross'))['value_gross__sum'] or 0
            expenses_net_sum = qs.filter(
                transaction_type=WITHDRAWAL).aggregate(
                    Sum('amount_net'))['amount_net__sum'] or 0
            expenses_gross_sum = qs.filter(
                transaction_type=WITHDRAWAL).aggregate(
                    Sum('amount_gross'))['amount_gross__sum'] or 0
            income_net_sum = qs.filter(transaction_type=DEPOSIT).aggregate(
                Sum('amount_net'))['amount_net__sum'] or 0
            income_gross_sum = qs.filter(transaction_type=DEPOSIT).aggregate(
                Sum('amount_gross'))['amount_gross__sum'] or 0

            amount_net_sum_base = amount_net_sum * rate
            amount_gross_sum_base = amount_gross_sum * rate
            expenses_net_sum_base = expenses_net_sum * rate
            expenses_gross_sum_base = expenses_gross_sum * rate
            income_net_sum_base = income_net_sum * rate
            income_gross_sum_base = income_gross_sum * rate

            totals['amount_net'] += amount_net_sum_base
            totals['amount_gross'] += amount_gross_sum_base
            totals['expenses_net'] += expenses_net_sum_base
            totals['expenses_gross'] += expenses_gross_sum_base
            totals['income_net'] += income_net_sum_base
            totals['income_gross'] += income_gross_sum_base

            account_transactions.append({
                'account': account,
                'account_balance': account_balance,
                'transactions': qs,
                'amount_net_total': amount_net_sum,
                'amount_gross_total': amount_gross_sum,
                'expenses_net_total': expenses_net_sum,
                'expenses_gross_total': expenses_gross_sum,
                'income_net_total': income_net_sum,
                'income_gross_total': income_gross_sum,
                'amount_net_sum_base': amount_net_sum_base,
                'amount_gross_sum_base': amount_gross_sum_base,
                'expenses_net_sum_base': expenses_net_sum_base,
                'expenses_gross_sum_base': expenses_gross_sum_base,
                'income_net_sum_base': income_net_sum_base,
                'income_gross_sum_base': income_gross_sum_base,
            })

        # We want to show a table: For each currency, show outstanding
        # income and expenses and profit. We also want a Total column. Of
        # course we need to convert all currency values to a base currency so
        # that we can create a sum for the total column.
        qs = self.get_outstanding_invoices()
        outstanding_expenses_gross_total_base = 0
        outstanding_income_gross_total_base = 0
        outstanding_profit_gross_total_base = 0
        outstanding_ccy_totals = {}
        for currency in Currency.objects.all():
            rate = 1
            if not currency.iso_code == base_currency:
                rate = self.get_rate(currency)
            outstanding_expenses_gross_sum = qs.filter(
                invoice_type=WITHDRAWAL, currency=currency).aggregate(
                    Sum('amount_gross'))['amount_gross__sum'] or 0
            outstanding_expenses_gross_sum_base = \
                outstanding_expenses_gross_sum * rate
            outstanding_expenses_gross_total_base += \
                outstanding_expenses_gross_sum_base

            outstanding_income_gross_sum = qs.filter(
                invoice_type=DEPOSIT, currency=currency).aggregate(
                    Sum('amount_gross'))['amount_gross__sum'] or 0
            outstanding_income_gross_sum_base = \
                outstanding_income_gross_sum * rate
            outstanding_income_gross_total_base += \
                outstanding_income_gross_sum_base

            outstanding_profit_gross_sum = \
                outstanding_income_gross_sum \
                - outstanding_expenses_gross_sum
            outstanding_profit_gross_sum_base = \
                outstanding_income_gross_sum_base \
                - outstanding_expenses_gross_sum_base
            outstanding_profit_gross_total_base += \
                outstanding_profit_gross_sum_base

            outstanding_ccy_totals[currency] = {
                'expenses_gross': outstanding_expenses_gross_sum,
                'expenses_gross_base': outstanding_expenses_gross_sum_base,
                'income_gross': outstanding_income_gross_sum,
                'income_gross_base': outstanding_income_gross_sum_base,
                'profit_gross': outstanding_profit_gross_sum,
                'profit_gross_base': outstanding_profit_gross_sum_base,
            }

        totals['outstanding_expenses_gross'] = \
            outstanding_expenses_gross_total_base
        totals['outstanding_income_gross'] = \
            outstanding_income_gross_total_base
        totals['outstanding_profit_gross'] = \
            outstanding_profit_gross_total_base

        ctx.update({
            'view_name': self.get_view_name(),
            'transaction_types': models.Transaction.TRANSACTION_TYPES,
            'base_currency': base_currency,
            'totals': totals,
            'account_transactions': account_transactions,
            'outstanding_invoices': qs,
            'outstanding_ccy_totals': outstanding_ccy_totals,
        })
        return ctx

    def get_view_name(self):
        """
        Return the name of the view that is displayed in the main h1-tag.

        """
        raise NotImplementedError('Method not implemented')  # pragma: no cover

    def get_outstanding_invoices(self):
        """
        Returns the transactions that should be shown in this view.

        """
        raise NotImplementedError('Method not implemented')  # pragma: no cover

    def get_rate(self, currency):
        """
        Returns the conversion rate.

        For month views, it should use a ConversionRate object and a filter
        on the given month.

        For other views, I'm not sure what to use.

        """
        raise NotImplementedError('Method not implemented')  # pragma: no cover

    def get_transactions(self, account):
        """
        Returns the transactions that should be shown in this view.

        """
        raise NotImplementedError('Method not implemented')  # pragma: no cover


class AllTimeView(AccountsViewMixin, generic.TemplateView):
    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.month = date(date.today().year, date.today().month, 1)
        return super(AllTimeView, self).dispatch(request, *args, **kwargs)

    def get_view_name(self):
        return 'All Time Overview'

    def get_outstanding_invoices(self):
        return models.Invoice.objects.filter(
            payment_date__isnull=True
        )

    def get_rate(self, currency):
        return decimal.Decimal(CurrencyRateHistory.objects.filter(
            rate__from_currency=currency,
            rate__to_currency__iso_code=getattr(
                settings, 'BASE_CURRENCY', 'EUR'),
        )[0].value)

    def get_transactions(self, account):
        return models.Transaction.objects.filter(
            account=account,
            parent__isnull=True,
        )


class CurrentMonthRedirectView(generic.View):
    """Redirects to the ``MonthOverviewView`` for the current month."""
    def dispatch(self, request, *args, **kwargs):
        now_ = now()
        return redirect(
            'account_keeping_month', year=now_.year, month=now_.month)


class CurrentYearRedirectView(generic.View):
    """Redirects to the ``MonthView`` for the current year."""
    def dispatch(self, request, *args, **kwargs):
        now_ = now()
        return redirect('account_keeping_year', year=now_.year)


class IndexView(LoginRequiredMixin, generic.TemplateView):
    """View that shows the main menu for the accounting app."""
    template_name = 'account_keeping/index_view.html'

    def get_context_data(self, **kwargs):
        ctx = super(IndexView, self).get_context_data(**kwargs)
        ctx.update({
            'invoices_without_pdf': models.Invoice.objects.get_without_pdf(),
            'transactions_without_invoice':
                models.Transaction.objects.get_without_invoice(),
            'transaction_types': models.Transaction.TRANSACTION_TYPES,
            'unpaid_invoices_with_transactions':
                get_unpaid_invoices_with_transactions(),
        })
        return ctx


class MonthView(AccountsViewMixin, generic.TemplateView):
    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.month = date(int(kwargs.get('year')), int(kwargs.get('month')), 1)
        return super(MonthView, self).dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super(MonthView, self).get_context_data(**kwargs)
        last_month = \
            self.month - relativedelta.relativedelta(months=1)
        next_month = \
            self.month + relativedelta.relativedelta(months=1)
        if next_month > date.today():
            next_month = None

        ctx.update({
            'month': self.month,
            'last_month': last_month,
            'next_month': next_month,
        })
        return ctx

    def get_view_name(self):
        return date_filter(self.month, 'F Y')

    def get_outstanding_invoices(self):
        next_month = self.month + relativedelta.relativedelta(months=1)
        return models.Invoice.objects.filter(
            Q(invoice_date__lt=next_month),
            Q(payment_date__isnull=True) | Q(payment_date__gte=next_month),
        ).prefetch_related('transactions')

    def get_rate(self, currency):
        rates = CurrencyRateHistory.objects.filter(
            rate__from_currency=currency,
            rate__to_currency__iso_code=getattr(
                settings, 'BASE_CURRENCY', 'EUR'),
        )
        try:
            return decimal.Decimal(rates.filter(
                date__year=self.month.year,
                date__month=self.month.month,
            )[0].value)
        except IndexError:
            # Get latest rate history record
            return decimal.Decimal(rates[0].value)

    def get_transactions(self, account):
        return models.Transaction.objects.filter(
            account=account,
            parent__isnull=True,
            transaction_date__year=self.month.year,
            transaction_date__month=self.month.month,
        )


class YearOverviewView(generic.TemplateView):
    template_name = 'account_keeping/year_view.html'

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.year = int(kwargs.get('year'))
        return super(YearOverviewView, self).dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super(YearOverviewView, self).get_context_data(**kwargs)
        past_months_of_year = utils.get_months_of_year(self.year)
        last_year = self.year - 1
        next_year = self.year + 1
        if next_year > date.today().year:
            next_year = None

        truncate_date = connection.ops.date_trunc_sql(
            'month', 'transaction_date')

        qs_year = models.Transaction.objects.filter(
            parent__isnull=True, transaction_date__year=self.year).extra(
                {'month': truncate_date})

        qs_income = qs_year.filter(
            transaction_type=DEPOSIT).values(
                'month', 'currency').annotate(
                    Sum('amount_gross')).order_by('currency', 'month')

        months = []
        for i in range(1, 13):
            months.append(date(self.year, i, 1))

        month_rates = {}
        base_currency = getattr(settings, 'BASE_CURRENCY', 'EUR')
        for month in months:
            for currency in Currency.objects.all():
                rate = 1
                if not currency.iso_code == base_currency:
                    rates = CurrencyRateHistory.objects.filter(
                        rate__from_currency=currency,
                        rate__to_currency__iso_code=base_currency,
                    )
                    try:
                        rate = decimal.Decimal(rates.filter(
                            date__year=month.year,
                            date__month=month.month,
                        )[0].value)
                    except IndexError:
                        # Get latest rate history record
                        rate = decimal.Decimal(rates[0].value)
                if month not in month_rates:
                    month_rates[month] = {}
                month_rates[month][currency.pk] = decimal.Decimal(rate)

        for row in qs_income:
            row['month'] = d(row['month']).date()
            row['amount_gross__sum'] = \
                row['amount_gross__sum'] \
                * month_rates[row['month']][row['currency']]

        income_total = {}
        for row in qs_income:
            if row['month'] not in income_total:
                income_total[row['month']] = 0
            income_total[row['month']] += row['amount_gross__sum']

        qs_expenses = qs_year.filter(
            transaction_type=WITHDRAWAL).values(
                'month', 'currency').annotate(
                    Sum('amount_gross')).order_by('currency', 'month')

        for row in qs_expenses:
            row['month'] = d(row['month']).date()
            row['amount_gross__sum'] = \
                row['amount_gross__sum'] \
                * month_rates[row['month']][row['currency']]

        expenses_total = {}
        for row in qs_expenses:
            if row['month'] not in expenses_total:
                expenses_total[row['month']] = 0
            expenses_total[row['month']] += row['amount_gross__sum']

        profit_total = {}
        for month in months:
            try:
                profit_total[month] = \
                    income_total[month] - expenses_total[month]
            except KeyError:
                pass

        # This adds a new column to the query which only holds the month
        # of the invoice_date and payment_date
        truncate_invoice_date = connection.ops.date_trunc_sql(
            'month', 'invoice_date')
        truncate_payment_date = connection.ops.date_trunc_sql(
            'payment_month', 'payment_date')
        qs_invoices_year = models.Invoice.objects.filter(
            invoice_date__year=self.year).extra({
                'month': truncate_invoice_date,
                'payment_month': truncate_payment_date})

        qs_new = qs_invoices_year.filter(invoice_type=DEPOSIT)
        qs_new = qs_new.values('month', 'currency')
        qs_new = qs_new.annotate(Sum('value_gross'))
        qs_new = qs_new.order_by('currency', 'month')

        for row in qs_new:
            row['month'] = d(row['month']).date()
            row['value_gross__sum'] = \
                row['value_gross__sum'] \
                * month_rates[row['month']][row['currency']]

        new_total = {}
        for row in qs_new:
            if row['month'] not in new_total:
                new_total[row['month']] = 0
            new_total[row['month']] += row['value_gross__sum']
        for month in months:
            if month not in new_total:
                new_total[month] = 0

        outstanding_total = {}
        for month in months:
            # for each month, we want to Sum the invoices, that were still
            # outstanding as of that month. An invoice sent in February and
            # paid in May would appear as outstanding on the months February,
            # March, April.
            if month.month > past_months_of_year:
                # In the current year, we don't want to save values for future
                # months
                break

            next_month = month + relativedelta.relativedelta(months=1)

            # TODO: Centralise this, we have this select in
            # get_outstanding_invoices already
            qs = models.Invoice.objects.filter(
                Q(invoice_date__lt=next_month),
                Q(payment_date__isnull=True) | Q(payment_date__gte=next_month))
            qs = qs.prefetch_related(
                'transactions')
            qs = qs.extra({'month': truncate_invoice_date, })
            qs = qs.values('currency')
            qs = qs.annotate(Sum('value_gross'))
            qs_outstanding_month = qs.order_by('currency')

            for row in qs_outstanding_month:
                row['value_gross__sum'] = \
                    row['value_gross__sum'] \
                    * month_rates[month][row['currency']]
                try:
                    outstanding_total[month] += row['value_gross__sum']
                except KeyError:
                    outstanding_total[month] = row['value_gross__sum']

        balance_total = {}
        for month in months:
            if month.month > past_months_of_year:
                # In the current year, we don't want to save values for future
                # months
                break

            month_end = month + relativedelta.relativedelta(
                months=1, seconds=-1)
            for account in models.Account.objects.all():
                qs_balance = models.Transaction.objects.filter(
                    account=account,
                    parent__isnull=True,
                    transaction_date__lte=month_end,
                ).aggregate(Sum('value_gross'))
                try:
                    qs_balance['value_gross__sum'] += account.initial_amount
                except TypeError:
                    # happens when there are no Transactions for that month
                    qs_balance['value_gross__sum'] = 0
                    qs_balance['value_gross__sum'] += account.initial_amount
                if not account.currency.iso_code == base_currency:
                    qs_balance['value_gross__sum'] = \
                        qs_balance['value_gross__sum'] \
                        * month_rates[month][account.currency.pk]
                if month not in balance_total:
                    balance_total[month] = 0
                balance_total[month] += qs_balance['value_gross__sum']

        equity_total = {}
        income_total_total = 0
        for month in months:
            try:
                equity_total[month] = \
                    balance_total[month] + outstanding_total[month]
            except KeyError:
                break

        for month in months:
            try:
                income_total_total += income_total[month]
            except KeyError:
                break
        income_average = income_total_total / past_months_of_year

        expenses_total_total = 0
        for month in months:
            try:
                expenses_total_total += expenses_total[month]
            except KeyError:
                break
        expenses_average = expenses_total_total / past_months_of_year

        profit_total_total = 0
        for month in months:
            try:
                profit_total_total += profit_total[month]
            except KeyError:
                break
        profit_average = profit_total_total / past_months_of_year

        new_total_total = 0
        for month in months:
            try:
                new_total_total += new_total[month]
            except KeyError:  # pragma: nocover
                break
        new_average = new_total_total / past_months_of_year

        outstanding_total_total = 0
        for month in months:
            try:
                outstanding_total_total += outstanding_total[month]
            except KeyError:
                break
        outstanding_average = outstanding_total_total / past_months_of_year

        balance_total_total = 0
        for month in months:
            try:
                balance_total_total += balance_total[month]
            except KeyError:
                break
        balance_average = balance_total_total / past_months_of_year

        equity_total_total = 0
        for month in months:
            try:
                equity_total_total += equity_total[month]
            except KeyError:
                break
        equity_average = equity_total_total / past_months_of_year

        ctx.update({
            'year': self.year,
            'last_year': last_year,
            'next_year': next_year,
            'months': months,
            'income_total': income_total,
            'expenses_total': expenses_total,
            'profit_total': profit_total,
            'new_total': new_total,
            'outstanding_total': outstanding_total,
            'balance_total': balance_total,
            'equity_total': equity_total,
            'income_total_total': income_total_total,
            'expenses_total_total': expenses_total_total,
            'profit_total_total': profit_total_total,
            'new_total_total': new_total_total,
            'income_average': income_average,
            'expenses_average': expenses_average,
            'profit_average': profit_average,
            'new_average': new_average,
            'outstanding_average': outstanding_average,
            'balance_average': balance_average,
            'equity_average': equity_average,
        })
        return ctx


class PayeeMixin(LoginRequiredMixin):
    model = models.Payee
    fields = '__all__'
    success_url = reverse_lazy('account_keeping_payees')


class PayeeListView(PayeeMixin, generic.ListView):
    pass


class PayeeCreateView(PayeeMixin, generic.CreateView):
    pass


class PayeeUpdateView(PayeeMixin, generic.UpdateView):
    pass


class AccountListView(LoginRequiredMixin, generic.ListView):
    model = models.Account


class InvoiceMixin(object):
    model = models.Invoice
    form_class = forms.InvoiceForm

    def get_success_url(self):
        return reverse('account_keeping_month', kwargs={
            'year': self.object.invoice_date.year,
            'month': self.object.invoice_date.month,
        })


class InvoiceCreateView(InvoiceMixin, LoginRequiredMixin, generic.CreateView):
    initial = {'invoice_date': now()}


class InvoiceUpdateView(InvoiceMixin, LoginRequiredMixin, generic.UpdateView):
    pass


class TransactionMixin(object):
    model = models.Transaction
    form_class = forms.TransactionForm

    def get_success_url(self):
        return u'{}#{}'.format(reverse('account_keeping_month', kwargs={
            'year': self.object.transaction_date.year,
            'month': self.object.transaction_date.month,
        }), self.object.account.slug)


class TransactionCreateView(TransactionMixin, LoginRequiredMixin,
                            generic.CreateView):
    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.initial.update({'invoice': self.request.GET.get('invoice')})
        self.initial.update({'parent': self.request.GET.get('parent')})
        return super(TransactionCreateView, self).dispatch(
            request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super(TransactionCreateView, self).get_form_kwargs()
        kwargs['initial'].update({
            'transaction_date': now(),
        })
        return kwargs


class TransactionUpdateView(TransactionMixin, LoginRequiredMixin,
                            generic.UpdateView):
    pass


class TransactionExportView(LoginRequiredMixin, generic.FormView):
    """Creates a csv, which includes a specific set of transactions."""
    template_name = 'account_keeping/export.html'
    form_class = forms.ExportForm

    def get_form_kwargs(self):
        kwargs = super(TransactionExportView, self).get_form_kwargs()
        kwargs['initial'].update({
            'start': date(now().today().year - 1, 1, 1),
            'end': date(now().today().year - 1, 12, 31),
        })
        return kwargs

    def form_valid(self, form):
        dataset = models.TransactionResource().export(
            queryset=models.Transaction.objects.filter(
                transaction_date__gte=form.cleaned_data['start'],
                transaction_date__lte=form.cleaned_data['end'],
            ).order_by('-transaction_date', '-pk')
        )
        response = HttpResponse(dataset.xls, content_type="application/csv")
        response['Content-Disposition'] = \
            u'attachment; filename="{} {} - {}.xls"'.format(
                form.cleaned_data['account'].name,
                form.cleaned_data['start'],
                form.cleaned_data['end'])
        return response



"""Tests for the forms of the account_keeping app."""
from django.test import TestCase
from django.utils.timezone import now

from mixer.backend.django import mixer

from .. import forms


class InvoiceFormTestCase(TestCase):
    """Tests for the ``InvoiceForm`` form."""
    longMessage = True

    def test_form(self):
        form = forms.InvoiceForm(data={
            'invoice_type': 'd',
            'invoice_date': now().strftime('%Y-%m-%d'),
            'currency': mixer.blend('currency_history.Currency').pk,
            'vat': 0,
            'value_net': 0,
            'value_gross': 0,
        })
        self.assertFalse(form.errors)


class TransactionFormTestCase(TestCase):
    """Tests for the ``TransactionForm`` form."""
    longMessage = True

    def test_form(self):
        data = {
            'transaction_type': 'd',
            'transaction_date': now().strftime('%Y-%m-%d'),
            'account': mixer.blend('account_keeping.Account').pk,
            'payee': mixer.blend('account_keeping.Payee').pk,
            'category': mixer.blend('account_keeping.Category').pk,
            'currency': mixer.blend('currency_history.Currency').pk,
            'amount_net': 0,
            'amount_gross': 0,
            'vat': 0,
            'value_net': 0,
            'value_gross': 0,
            'mark_invoice': True,
        }
        form = forms.TransactionForm(data=data)
        self.assertFalse(form.errors)
        transaction = form.save()
        transaction.invoice = mixer.blend('account_keeping.Invoice',
                                          payment_date=None)
        transaction.invoice.save()
        self.assertFalse(transaction.invoice.payment_date)
        data.update({'invoice': transaction.invoice.pk})
        form = forms.TransactionForm(data=data)
        self.assertFalse(form.errors)
        transaction = form.save()
        self.assertTrue(transaction.invoice.payment_date)



"""API calls against letsfreckle.com"""
from django.conf import settings
from django.utils.translation import ugettext_lazy as _

try:
    from freckle_client.client import FreckleClientV2
    client = FreckleClientV2(settings.ACCOUNT_KEEPING_FRECKLE_ACCESS_TOKEN)
except ImportError:  # pragma: nocover
    client = None
from requests.exceptions import ConnectionError, HTTPError

from . import models


def get_unpaid_invoices_with_transactions():
    """
    Returns all invoices that are unpaid on freckle but have transactions.

    This means, that the invoice is either partially paid and can be left as
    unpaid in freckle, or the invoice has been fully paid and should be set to
    paid in freckle as well.

    """
    if not client:  # pragma: nocover
        return None
    result = {}
    try:
        unpaid_invoices = client.fetch_json(
            'invoices', query_params={'state': 'unpaid'})
    except (ConnectionError, HTTPError):  # pragma: nocover
        result.update({'error': _('Wasn\'t able to connect to Freckle.')})
    else:
        invoices = []
        for invoice in unpaid_invoices:
            invoice_with_transactions = models.Invoice.objects.filter(
                invoice_number=invoice['reference'],
                transactions__isnull=False)
            if invoice_with_transactions:
                invoices.append(invoice)
        result.update({'invoices': invoices})
    return result



"""
Models for the account_keeping app.

TODO: Add lazy_trans and docstrings

"""
from datetime import date
from decimal import Decimal

from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from currency_history.models import Currency, CurrencyRateHistory
from dateutil import relativedelta
from import_export import resources
from import_export.fields import Field


class AmountMixin(object):
    """
    Mixin that handles amount_net, vat and amount_gross fields on save().

    """
    def set_amount_fields(self):
        if self.amount_net and not self.amount_gross:
            if self.vat:
                self.amount_gross = \
                    self.amount_net * (self.vat / Decimal(100.0) + 1)
            else:
                self.amount_gross = self.amount_net

        if self.amount_gross and not self.amount_net:
            if self.vat:
                self.amount_net = \
                    Decimal(1.0) / (self.vat / Decimal(100.0) + 1) \
                    * self.amount_gross
            else:
                self.amount_net = self.amount_gross

    def set_value_fields(self, type_field_name):
        multiplier = 1
        type_ = getattr(self, type_field_name)
        if type_ == Transaction.TRANSACTION_TYPES['withdrawal']:
            multiplier = -1
        self.value_net = self.amount_net * multiplier
        self.value_gross = self.amount_gross * multiplier


@python_2_unicode_compatible
class Account(models.Model):
    name = models.CharField(
        max_length=128,
        verbose_name=_('Name'),
    )

    slug = models.SlugField(
        max_length=128,
        verbose_name=_('Slug'),
    )

    currency = models.ForeignKey(
        'currency_history.Currency',
        related_name='accounts',
        verbose_name=_('Currency'),
    )

    initial_amount = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0,
        verbose_name=_('Initial amount'),
    )

    total_amount = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0,
        verbose_name=_('Total amount'),
    )

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']
        verbose_name = _('Account')
        verbose_name_plural = _('Accounts')

    def get_balance(self, month=None):
        """
        Returns the balance up until now or until the provided month.

        """
        if not month:
            month = date(date.today().year, date.today().month, 1)
        next_month = month + relativedelta.relativedelta(months=1)
        account_balance = self.transactions.filter(
            parent__isnull=True,
            transaction_date__lt=next_month,
        ).aggregate(models.Sum('value_gross'))['value_gross__sum'] or 0
        account_balance = account_balance + self.initial_amount
        return account_balance


class InvoiceManager(models.Manager):
    """Custom manager for the ``Invoice`` model."""
    def get_without_pdf(self):
        qs = Invoice.objects.filter(pdf='')
        qs = qs.prefetch_related('transactions', )
        return qs


@python_2_unicode_compatible
class Invoice(AmountMixin, models.Model):
    INVOICE_TYPES = {
        'withdrawal': 'w',
        'deposit': 'd',
    }

    INVOICE_TYPE_CHOICES = [
        (INVOICE_TYPES['withdrawal'], 'withdrawal'),
        (INVOICE_TYPES['deposit'], 'deposit'),
    ]

    invoice_type = models.CharField(
        max_length=1,
        choices=INVOICE_TYPE_CHOICES,
        verbose_name=_('Invoice type'),
    )

    invoice_date = models.DateField(
        verbose_name=_('Invoice date'),
    )

    invoice_number = models.CharField(
        max_length=256,
        verbose_name=_('Invoice No.'),
        blank=True,
    )

    description = models.TextField(
        verbose_name=_('Description'),
        blank=True,
    )

    currency = models.ForeignKey(
        'currency_history.Currency',
        related_name='invoices',
        verbose_name=_('Currency'),
    )

    amount_net = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0,
        blank=True,
        verbose_name=_('Amount net'),
    )

    vat = models.DecimalField(
        max_digits=4,
        decimal_places=2,
        default=0,
        verbose_name=_('VAT'),
    )

    amount_gross = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0,
        blank=True,
        verbose_name=_('Amount gross'),
    )

    value_net = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0,
        verbose_name=_('Value net'),
    )

    value_gross = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0,
        verbose_name=_('Value gross'),
    )

    payment_date = models.DateField(
        verbose_name=_('Payment date'),
        blank=True,
        null=True,
    )

    pdf = models.FileField(
        upload_to='invoice_files',
        verbose_name=_('PDF'),
        blank=True, null=True,
    )

    objects = InvoiceManager()

    class Meta:
        ordering = ['-invoice_date', '-pk']
        verbose_name = _('Invoice')
        verbose_name_plural = _('Invoices')

    def __str__(self):
        if self.invoice_number:
            return self.invoice_number
        return '{0} - {1}'.format(self.invoice_date,
                                  self.get_invoice_type_display())

    def save(self, *args, **kwargs):
        self.set_amount_fields()
        self.set_value_fields('invoice_type')
        return super(Invoice, self).save(*args, **kwargs)

    @property
    def balance(self):
        if not self.transactions.all():
            return 0 - self.amount_net

        total = 0
        # Convert amounts
        for currency in Currency.objects.all():
            # Get transactions for each currency
            transactions = self.transactions.filter(currency=currency)
            if not transactions:
                continue

            if currency == self.currency:
                rate = 1
            else:
                rate = Decimal(CurrencyRateHistory.objects.filter(
                    rate__from_currency=currency,
                    rate__to_currency=self.currency,
                )[0].value)
            total += rate * transactions.aggregate(
                models.Sum('amount_net'))['amount_net__sum']
        return total - self.amount_net


@python_2_unicode_compatible
class Payee(models.Model):
    name = models.CharField(
        verbose_name=_('Payee'),
        max_length=256,
    )

    class Meta:
        ordering = ['name', ]
        verbose_name = _('Payee')
        verbose_name_plural = _('Payees')

    def __str__(self):
        return self.name

    def invoices(self):
        return Invoice.objects.filter(
            pk__in=self.transactions.values_list('invoice__pk')).distinct()


@python_2_unicode_compatible
class Category(models.Model):
    name = models.CharField(
        max_length=256,
        verbose_name=_('Name'),
    )

    class Meta:
        ordering = ['name', ]
        verbose_name = _('Category')
        verbose_name_plural = _('Categories')

    def __str__(self):
        return self.name


class TransactionManager(models.Manager):
    """Manager for the ``Transaction`` model."""
    def get_totals_by_payee(self, account, start_date=None, end_date=None):
        """
        Returns transaction totals grouped by Payee.

        """
        qs = Transaction.objects.filter(account=account, parent__isnull=True)
        qs = qs.values('payee').annotate(models.Sum('value_gross'))
        qs = qs.order_by('payee__name')
        return qs

    def get_without_invoice(self):
        """
        Returns transactions that don't have an invoice.

        We filter out transactions that have children, because those
        transactions never have invoices - their children are the ones that
        would each have one invoice.

        """
        qs = Transaction.objects.filter(
            children__isnull=True, invoice__isnull=True)
        return qs


@python_2_unicode_compatible
class Transaction(AmountMixin, models.Model):
    TRANSACTION_TYPES = {
        'withdrawal': 'w',
        'deposit': 'd',
    }

    TRANSACTION_TYPE_CHOICES = [
        (TRANSACTION_TYPES['withdrawal'], 'withdrawal'),
        (TRANSACTION_TYPES['deposit'], 'deposit'),
    ]

    account = models.ForeignKey(
        Account,
        related_name='transactions',
        verbose_name=_('Account'),
    )

    parent = models.ForeignKey(
        'account_keeping.Transaction',
        related_name='children',
        blank=True, null=True,
        verbose_name=_('Parent'),
    )

    transaction_type = models.CharField(
        max_length=1,
        choices=TRANSACTION_TYPE_CHOICES,
        verbose_name=_('Transaction type'),
    )

    transaction_date = models.DateField(
        verbose_name=_('Transaction date'),
    )

    description = models.TextField(
        verbose_name=_('Description'),
        blank=True,
    )

    invoice_number = models.CharField(
        verbose_name=_('Invoice No.'),
        max_length=256,
        blank=True,
    )

    invoice = models.ForeignKey(
        Invoice,
        blank=True, null=True,
        related_name='transactions',
        verbose_name=_('Invoice'),
    )

    payee = models.ForeignKey(
        Payee,
        related_name='transactions',
        verbose_name=_('Payee'),
    )

    category = models.ForeignKey(
        Category,
        related_name='transactions',
        verbose_name=_('Category'),
    )

    currency = models.ForeignKey(
        'currency_history.Currency',
        related_name='transactions',
        verbose_name=_('Currency'),
    )

    amount_net = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0,
        blank=True,
        verbose_name=_('Amount net'),
    )

    vat = models.DecimalField(
        max_digits=4,
        decimal_places=2,
        default=0,
        verbose_name=_('VAT'),
    )

    amount_gross = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0,
        blank=True,
        verbose_name=_('Amount gross'),
    )

    value_net = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0,
        verbose_name=_('Value net'),
    )

    value_gross = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0,
        verbose_name=_('Value gross'),
    )

    objects = TransactionManager()

    class Meta:
        ordering = ['-transaction_date', '-pk']
        verbose_name = _('Transaction')
        verbose_name_plural = _('Transactions')

    def __str__(self):
        if self.invoice_number:
            return self.invoice_number
        if self.invoice and self.invoice.invoice_number:
            return self.invoice.invoice_number
        return '{0} - {1}'.format(self.payee, self.category)

    def get_description(self):
        if self.description:
            return self.description
        if self.invoice and self.invoice.description:
            return self.invoice.description
        description = ''
        for child in self.children.all():
            if child.description:
                description += u'{0},\n'.format(child.description)
            elif child.invoice and child.invoice.description:
                description += u'{0},\n'.format(child.invoice.description)
        return description or u'n/a'

    def get_invoices(self):
        if self.children.all():
            return [child.invoice for child in self.children.all()]
        return [self.invoice, ]

    def save(self, *args, **kwargs):
        self.set_amount_fields()
        self.set_value_fields('transaction_type')
        return super(Transaction, self).save(*args, **kwargs)


class TransactionResource(resources.ModelResource):
    invoice_no = Field(column_name='Invoice No.')
    balance = Field(column_name='Balance')
    get_transaction_type = Field(column_name='Transaction type')

    @classmethod
    def field_from_django_field(self, field_name, django_field, readonly):
        field = resources.ModelResource.field_from_django_field(
            field_name, django_field, readonly)
        field.column_name = django_field.verbose_name
        return field

    class Meta:
        model = Transaction
        fields = ('id', 'transaction_date', 'description', 'invoice_no',
                  'payee__name', 'category__name', 'get_transaction_type',
                  'currency__iso_code', 'amount_net', 'vat', 'amount_gross',
                  'balance')
        export_order = fields

    def dehydrate_invoice_no(self, transaction):  # pragma: nocover
        if transaction.invoice_number:
            return transaction.invoice_number
        if transaction.invoice and transaction.invoice.invoice_number:
            return transaction.invoice.invoice_number
        return ''

    def dehydrate_balance(self, transaction):  # pragma: nocover
        account_balance = transaction.account.transactions.filter(
            models.Q(parent__isnull=True),
            models.Q(transaction_date__lt=transaction.transaction_date) |
            (models.Q(transaction_date=transaction.transaction_date,
                      pk__lte=transaction.pk)),
        ).aggregate(models.Sum('value_gross'))['value_gross__sum'] or 0
        return account_balance + transaction.account.initial_amount

    def dehydrate_get_transaction_type(self, transaction):  # pragma: nocover
        return transaction.get_transaction_type_display()



from django.conf import settings
from django import template
from django.contrib.humanize.templatetags.humanize import intcomma

register = template.Library()


@register.filter()
def currency(value, currency=None):
    try:
        currency = currency.iso_code
    except AttributeError:
        currency = getattr(settings, 'BASE_CURRENCY', 'EUR')
    cleaned_value = '{0:.2f}'.format(float(value))
    return u'{0} {1}'.format(currency, intcomma(cleaned_value))



"""Tests for the models of the account_keeping app."""
from django.test import TestCase
from django.utils.timezone import now, timedelta

from mixer.backend.django import mixer

from .. import models


WITHDRAWAL = models.Transaction.TRANSACTION_TYPES['withdrawal']
DEPOSIT = models.Transaction.TRANSACTION_TYPES['deposit']


class AccountTestCase(TestCase):
    """Tests for the ``Account`` model."""
    def setUp(self):
        self.account = mixer.blend('account_keeping.Account')

    def test_model(self):
        self.assertTrue(str(self.account))

    def test_get_balance(self):
        self.assertEqual(self.account.get_balance(), 0)


class InvoiceTestCase(TestCase):
    """Tests for the ``Invoice`` model."""
    def test_model(self):
        obj = mixer.blend('account_keeping.Invoice', invoice_number='')
        self.assertTrue(str(obj))
        obj.invoice_number = 'Foo123'
        obj.save()
        self.assertEqual(str(obj), 'Foo123')

    def test_manager(self):
        mixer.blend('account_keeping.Invoice')
        mixer.blend('account_keeping.Invoice')
        self.assertEqual(models.Invoice.objects.get_without_pdf().count(), 2)

    def test_balance(self):
        invoice = mixer.blend('account_keeping.Invoice', amount_net=100)
        self.assertEqual(int(invoice.balance), -100)
        mixer.blend('account_keeping.Transaction', amount_net=10,
                    invoice=invoice, currency=invoice.currency)
        transaction = mixer.blend('account_keeping.Transaction', amount_net=50,
                                  invoice=invoice)
        mixer.blend('currency_history.CurrencyRateHistory', value=0.5,
                    rate__from_currency=transaction.currency,
                    rate__to_currency=invoice.currency)
        self.assertEqual(int(invoice.balance), -65)


class PayeeTestCase(TestCase):
    """Tests for the ``Payee`` model."""
    def setUp(self):
        self.payee = mixer.blend('account_keeping.Payee')

    def test_model(self):
        self.assertTrue(str(self.payee))

    def test_invoices(self):
        self.assertEqual(self.payee.invoices().count(), 0)


class CategoryTestCase(TestCase):
    """Tests for the ``Category`` model."""
    def test_model(self):
        obj = mixer.blend('account_keeping.Category')
        self.assertTrue(str(obj))


class TransactionTestCase(TestCase):
    """Tests for the ``Transaction`` model."""
    longMessage = True

    def setUp(self):
        self.transaction = mixer.blend('account_keeping.Transaction',
                                       invoice_number='')

    def test_model(self):
        self.assertTrue(str(self.transaction))
        self.transaction.invoice = mixer.blend('account_keeping.Invoice')
        self.assertEqual(str(self.transaction),
                         self.transaction.invoice.invoice_number)
        self.transaction.invoice_number = 'Foo123'
        self.assertEqual(str(self.transaction), 'Foo123')

    def test_save(self):
        obj = mixer.blend(
            'account_keeping.Transaction',
            amount_net=100, vat=19, amount_gross=None)
        self.assertEqual(obj.amount_gross, 119, msg=(
            'If only amount_net is given, amount_gross should be calculated'))

        obj = mixer.blend(
            'account_keeping.Transaction',
            amount_net=100, vat=0, amount_gross=None)
        self.assertEqual(obj.amount_gross, 100, msg=(
            'If only amount_net is given and VAT is 0, amount_gross should be'
            ' identical to amount_net'))

        obj = mixer.blend(
            'account_keeping.Transaction',
            amount_net=None, vat=19, amount_gross=119)
        self.assertEqual(obj.amount_net, 100, msg=(
            'If only amount_gross is given, amount_net should be calculated'))

        obj = mixer.blend(
            'account_keeping.Transaction',
            amount_net=None, vat=0, amount_gross=119)
        self.assertEqual(obj.amount_gross, 119, msg=(
            'If only amount_gross is given and VAT is 0, amount_net should be'
            ' identical to amount_gross'))

        obj = mixer.blend('account_keeping.Transaction',
                          transaction_type=DEPOSIT)
        self.assertEqual(obj.value_net, obj.amount_net, msg=(
            'When type is deposit, the value should be positive'))

        obj = mixer.blend('account_keeping.Transaction',
                          transaction_type=WITHDRAWAL)
        self.assertEqual(obj.value_net, obj.amount_net * -1, msg=(
            'When type is withdrawal, the value should be negative'))

    def test_get_description(self):
        """Tests for the ``get_description`` method."""
        trans = mixer.blend('account_keeping.Transaction', description='')
        result = trans.get_description()
        self.assertEqual(result, 'n/a', msg=(
            'If no description is set, it should return `n/a`'))

        trans.description = 'foo'
        trans.save()
        result = trans.get_description()
        self.assertEqual(result, 'foo', msg=(
            'If description is set, it should return the description'))

        invoice = mixer.blend('account_keeping.Invoice', description='barfoo')
        trans = mixer.blend('account_keeping.Transaction', invoice=invoice,
                            description='')
        result = trans.get_description()
        self.assertEqual(result, 'barfoo', msg=(
            'If no description is set but the invoice has one, it should'
            ' return that one'))

        trans = mixer.blend('account_keeping.Transaction', description='')
        mixer.blend('account_keeping.Transaction', parent=trans,
                    description='foofoo',
                    transaction_date=now() - timedelta(days=2))
        mixer.blend('account_keeping.Transaction', parent=trans,
                    description='barbar',
                    transaction_date=now() - timedelta(days=1))
        result = trans.get_description()
        self.assertEqual(result, 'barbar,\nfoofoo,\n', msg=(
            'If no description is set but children have one, it should'
            ' return the descriptions of the children'))

        trans = mixer.blend('account_keeping.Transaction', description='')
        mixer.blend('account_keeping.Transaction',
                    parent=trans, invoice=invoice, description='')
        mixer.blend('account_keeping.Transaction', parent=trans,
                    description='barbar',
                    transaction_date=now() - timedelta(days=1))
        result = trans.get_description()
        self.assertEqual(result, 'barbar,\nbarfoo,\n', msg=(
            'If no description is set and children are present, but a child'
            ' doesn\'t have a description, it should return the description of'
            ' the child\'s invoice'))

    def test_get_invoices(self):
        """Tests for the ``get_invoices`` method."""
        trans = mixer.blend('account_keeping.Transaction')
        result = trans.get_invoices()
        self.assertEqual(result, [None, ], msg=(
            'If it doesn`t have an invoice, it should return a list with'
            ' None'))

        invoice = mixer.blend('account_keeping.Invoice')
        trans.invoice = invoice
        trans.save
        result = trans.get_invoices()
        self.assertEqual(result, [invoice, ], msg=(
            'If it has an invoice, it should return a list with the invoice'))

        trans = mixer.blend('account_keeping.Transaction')
        mixer.blend('account_keeping.Transaction',
                    parent=trans, invoice=invoice)
        result = trans.get_invoices()
        self.assertEqual(result, [invoice, ], msg=(
            'If it does not have an invoice, it should return the invoices of'
            ' it`s children'))

    def test_get_totals_by_payee(self):
        """Tests for the ``get_totals_by_payee`` method."""
        trans1a = mixer.blend('account_keeping.Transaction', value_gross=11.9)
        mixer.blend('account_keeping.Transaction', value_gross=11.9,
                    account=trans1a.account, payee=trans1a.payee)
        mixer.blend('account_keeping.Transaction', account=trans1a.account,
                    value_gross=11.9)
        result = models.Transaction.objects.get_totals_by_payee(
            trans1a.account)
        self.assertEqual(result.count(), 2, msg=(
            'Should return all transactions grouped by payee'))

    def test_manager(self):
        self.assertEqual(
            models.Transaction.objects.get_without_invoice().count(), 1)



# flake8: noqa
# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Currency'
        db.create_table(u'account_keeping_currency', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=64)),
            ('iso_code', self.gf('django.db.models.fields.CharField')(max_length=3)),
            ('is_base_currency', self.gf('django.db.models.fields.BooleanField')(default=False)),
        ))
        db.send_create_signal(u'account_keeping', ['Currency'])

        # Adding model 'CurrencyRate'
        db.create_table(u'account_keeping_currencyrate', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('currency', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['account_keeping.Currency'])),
            ('year', self.gf('django.db.models.fields.PositiveIntegerField')()),
            ('month', self.gf('django.db.models.fields.PositiveIntegerField')()),
            ('rate', self.gf('django.db.models.fields.DecimalField')(max_digits=18, decimal_places=8)),
        ))
        db.send_create_signal(u'account_keeping', ['CurrencyRate'])

        # Adding model 'Account'
        db.create_table(u'account_keeping_account', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=128)),
            ('slug', self.gf('django.db.models.fields.SlugField')(max_length=128)),
            ('currency', self.gf('django.db.models.fields.related.ForeignKey')(related_name='accounts', to=orm['account_keeping.Currency'])),
            ('initial_amount', self.gf('django.db.models.fields.DecimalField')(default=0, max_digits=10, decimal_places=2)),
            ('total_amount', self.gf('django.db.models.fields.DecimalField')(default=0, max_digits=10, decimal_places=2)),
        ))
        db.send_create_signal(u'account_keeping', ['Account'])

        # Adding model 'Invoice'
        db.create_table(u'account_keeping_invoice', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('invoice_type', self.gf('django.db.models.fields.CharField')(max_length=1)),
            ('invoice_date', self.gf('django.db.models.fields.DateField')()),
            ('invoice_number', self.gf('django.db.models.fields.CharField')(max_length=256, blank=True)),
            ('currency', self.gf('django.db.models.fields.related.ForeignKey')(related_name='invoices', to=orm['account_keeping.Currency'])),
            ('amount_net', self.gf('django.db.models.fields.DecimalField')(default=0, max_digits=10, decimal_places=2)),
            ('vat', self.gf('django.db.models.fields.DecimalField')(default=0, max_digits=4, decimal_places=2)),
            ('amount_gross', self.gf('django.db.models.fields.DecimalField')(default=0, max_digits=10, decimal_places=2)),
            ('payment_date', self.gf('django.db.models.fields.DateField')(null=True, blank=True)),
            ('pdf', self.gf('django.db.models.fields.files.FileField')(max_length=100, null=True, blank=True)),
        ))
        db.send_create_signal(u'account_keeping', ['Invoice'])

        # Adding model 'Payee'
        db.create_table(u'account_keeping_payee', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=256)),
        ))
        db.send_create_signal(u'account_keeping', ['Payee'])

        # Adding model 'Category'
        db.create_table(u'account_keeping_category', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=256)),
        ))
        db.send_create_signal(u'account_keeping', ['Category'])

        # Adding model 'Transaction'
        db.create_table(u'account_keeping_transaction', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('account', self.gf('django.db.models.fields.related.ForeignKey')(related_name='transactions', to=orm['account_keeping.Account'])),
            ('parent', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='children', null=True, to=orm['account_keeping.Transaction'])),
            ('transaction_type', self.gf('django.db.models.fields.CharField')(max_length=1)),
            ('transaction_date', self.gf('django.db.models.fields.DateField')()),
            ('description', self.gf('django.db.models.fields.TextField')(blank=True)),
            ('invoice_number', self.gf('django.db.models.fields.CharField')(max_length=256, blank=True)),
            ('invoice', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='transactions', null=True, to=orm['account_keeping.Invoice'])),
            ('payee', self.gf('django.db.models.fields.related.ForeignKey')(related_name='transactions', to=orm['account_keeping.Payee'])),
            ('category', self.gf('django.db.models.fields.related.ForeignKey')(related_name='transactions', to=orm['account_keeping.Category'])),
            ('currency', self.gf('django.db.models.fields.related.ForeignKey')(related_name='transactions', to=orm['account_keeping.Currency'])),
            ('amount_net', self.gf('django.db.models.fields.DecimalField')(default=0, max_digits=10, decimal_places=2, blank=True)),
            ('vat', self.gf('django.db.models.fields.DecimalField')(default=0, max_digits=4, decimal_places=2)),
            ('amount_gross', self.gf('django.db.models.fields.DecimalField')(default=0, max_digits=10, decimal_places=2, blank=True)),
            ('value_net', self.gf('django.db.models.fields.DecimalField')(default=0, max_digits=10, decimal_places=2)),
            ('value_gross', self.gf('django.db.models.fields.DecimalField')(default=0, max_digits=10, decimal_places=2)),
        ))
        db.send_create_signal(u'account_keeping', ['Transaction'])


    def backwards(self, orm):
        # Deleting model 'Currency'
        db.delete_table(u'account_keeping_currency')

        # Deleting model 'CurrencyRate'
        db.delete_table(u'account_keeping_currencyrate')

        # Deleting model 'Account'
        db.delete_table(u'account_keeping_account')

        # Deleting model 'Invoice'
        db.delete_table(u'account_keeping_invoice')

        # Deleting model 'Payee'
        db.delete_table(u'account_keeping_payee')

        # Deleting model 'Category'
        db.delete_table(u'account_keeping_category')

        # Deleting model 'Transaction'
        db.delete_table(u'account_keeping_transaction')


    models = {
        u'account_keeping.account': {
            'Meta': {'object_name': 'Account'},
            'currency': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'accounts'", 'to': u"orm['account_keeping.Currency']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'initial_amount': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '128'}),
            'total_amount': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2'})
        },
        u'account_keeping.category': {
            'Meta': {'object_name': 'Category'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '256'})
        },
        u'account_keeping.currency': {
            'Meta': {'object_name': 'Currency'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_base_currency': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'iso_code': ('django.db.models.fields.CharField', [], {'max_length': '3'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '64'})
        },
        u'account_keeping.currencyrate': {
            'Meta': {'object_name': 'CurrencyRate'},
            'currency': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['account_keeping.Currency']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'month': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'rate': ('django.db.models.fields.DecimalField', [], {'max_digits': '18', 'decimal_places': '8'}),
            'year': ('django.db.models.fields.PositiveIntegerField', [], {})
        },
        u'account_keeping.invoice': {
            'Meta': {'ordering': "['invoice_date']", 'object_name': 'Invoice'},
            'amount_gross': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2'}),
            'amount_net': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2'}),
            'currency': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'invoices'", 'to': u"orm['account_keeping.Currency']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'invoice_date': ('django.db.models.fields.DateField', [], {}),
            'invoice_number': ('django.db.models.fields.CharField', [], {'max_length': '256', 'blank': 'True'}),
            'invoice_type': ('django.db.models.fields.CharField', [], {'max_length': '1'}),
            'payment_date': ('django.db.models.fields.DateField', [], {'null': 'True', 'blank': 'True'}),
            'pdf': ('django.db.models.fields.files.FileField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'vat': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '4', 'decimal_places': '2'})
        },
        u'account_keeping.payee': {
            'Meta': {'object_name': 'Payee'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '256'})
        },
        u'account_keeping.transaction': {
            'Meta': {'ordering': "['transaction_date']", 'object_name': 'Transaction'},
            'account': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'transactions'", 'to': u"orm['account_keeping.Account']"}),
            'amount_gross': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2', 'blank': 'True'}),
            'amount_net': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2', 'blank': 'True'}),
            'category': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'transactions'", 'to': u"orm['account_keeping.Category']"}),
            'currency': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'transactions'", 'to': u"orm['account_keeping.Currency']"}),
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'invoice': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'transactions'", 'null': 'True', 'to': u"orm['account_keeping.Invoice']"}),
            'invoice_number': ('django.db.models.fields.CharField', [], {'max_length': '256', 'blank': 'True'}),
            'parent': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'children'", 'null': 'True', 'to': u"orm['account_keeping.Transaction']"}),
            'payee': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'transactions'", 'to': u"orm['account_keeping.Payee']"}),
            'transaction_date': ('django.db.models.fields.DateField', [], {}),
            'transaction_type': ('django.db.models.fields.CharField', [], {'max_length': '1'}),
            'value_gross': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2'}),
            'value_net': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2'}),
            'vat': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '4', 'decimal_places': '2'})
        }
    }

    complete_apps = ['account_keeping']



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding field 'Invoice.value_net'
        db.add_column(u'account_keeping_invoice', 'value_net',
                      self.gf('django.db.models.fields.DecimalField')(default=0, max_digits=10, decimal_places=2),
                      keep_default=False)

        # Adding field 'Invoice.value_gross'
        db.add_column(u'account_keeping_invoice', 'value_gross',
                      self.gf('django.db.models.fields.DecimalField')(default=0, max_digits=10, decimal_places=2),
                      keep_default=False)


    def backwards(self, orm):
        # Deleting field 'Invoice.value_net'
        db.delete_column(u'account_keeping_invoice', 'value_net')

        # Deleting field 'Invoice.value_gross'
        db.delete_column(u'account_keeping_invoice', 'value_gross')


    models = {
        u'account_keeping.account': {
            'Meta': {'object_name': 'Account'},
            'currency': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'accounts'", 'to': u"orm['account_keeping.Currency']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'initial_amount': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'slug': ('django.db.models.fields.SlugField', [], {'max_length': '128'}),
            'total_amount': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2'})
        },
        u'account_keeping.category': {
            'Meta': {'ordering': "['name']", 'object_name': 'Category'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '256'})
        },
        u'account_keeping.currency': {
            'Meta': {'object_name': 'Currency'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_base_currency': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'iso_code': ('django.db.models.fields.CharField', [], {'max_length': '3'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '64'})
        },
        u'account_keeping.currencyrate': {
            'Meta': {'object_name': 'CurrencyRate'},
            'currency': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['account_keeping.Currency']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'month': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'rate': ('django.db.models.fields.DecimalField', [], {'max_digits': '18', 'decimal_places': '8'}),
            'year': ('django.db.models.fields.PositiveIntegerField', [], {})
        },
        u'account_keeping.invoice': {
            'Meta': {'ordering': "['invoice_date']", 'object_name': 'Invoice'},
            'amount_gross': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2', 'blank': 'True'}),
            'amount_net': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2', 'blank': 'True'}),
            'currency': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'invoices'", 'to': u"orm['account_keeping.Currency']"}),
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'invoice_date': ('django.db.models.fields.DateField', [], {}),
            'invoice_number': ('django.db.models.fields.CharField', [], {'max_length': '256', 'blank': 'True'}),
            'invoice_type': ('django.db.models.fields.CharField', [], {'max_length': '1'}),
            'payment_date': ('django.db.models.fields.DateField', [], {'null': 'True', 'blank': 'True'}),
            'pdf': ('django.db.models.fields.files.FileField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'value_gross': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2'}),
            'value_net': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2'}),
            'vat': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '4', 'decimal_places': '2'})
        },
        u'account_keeping.payee': {
            'Meta': {'ordering': "['name']", 'object_name': 'Payee'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '256'})
        },
        u'account_keeping.transaction': {
            'Meta': {'ordering': "['transaction_date']", 'object_name': 'Transaction'},
            'account': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'transactions'", 'to': u"orm['account_keeping.Account']"}),
            'amount_gross': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2', 'blank': 'True'}),
            'amount_net': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2', 'blank': 'True'}),
            'category': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'transactions'", 'to': u"orm['account_keeping.Category']"}),
            'currency': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'transactions'", 'to': u"orm['account_keeping.Currency']"}),
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'invoice': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'transactions'", 'null': 'True', 'to': u"orm['account_keeping.Invoice']"}),
            'invoice_number': ('django.db.models.fields.CharField', [], {'max_length': '256', 'blank': 'True'}),
            'parent': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'children'", 'null': 'True', 'to': u"orm['account_keeping.Transaction']"}),
            'payee': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'transactions'", 'to': u"orm['account_keeping.Payee']"}),
            'transaction_date': ('django.db.models.fields.DateField', [], {}),
            'transaction_type': ('django.db.models.fields.CharField', [], {'max_length': '1'}),
            'value_gross': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2'}),
            'value_net': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '10', 'decimal_places': '2'}),
            'vat': ('django.db.models.fields.DecimalField', [], {'default': '0', 'max_digits': '4', 'decimal_places': '2'})
        }
    }

    complete_apps = ['account_keeping']


"""Tests for the views of the account_keeping app."""
import json

from django.core.urlresolvers import reverse
from django.test import TestCase
from django.utils.timezone import now, timedelta

from django_libs.tests.mixins import ViewRequestFactoryTestMixin
from mixer.backend.django import mixer
from mock import patch
from requests import Response

from .. import views


class AllTimeViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``AllTimeView`` view class."""
    view_class = views.AllTimeView

    def setUp(self):
        self.user = mixer.blend('auth.User', is_superuser=True)
        self.ccy = mixer.blend('currency_history.Currency', iso_code='EUR')
        self.account = mixer.blend('account_keeping.Account',
                                   currency=self.ccy)
        self.trans1 = mixer.blend(
            'account_keeping.Transaction',
            account=self.account, currency=self.ccy)

        self.ccy2 = mixer.blend('currency_history.Currency')
        self.account2 = mixer.blend('account_keeping.Account',
                                    currency=self.ccy2)
        self.trans2 = mixer.blend(
            'account_keeping.Transaction',
            account=self.account2, currency=self.ccy2)
        mixer.blend(
            'currency_history.CurrencyRateHistory',
            rate__from_currency=self.ccy2,
            rate__to_currency=self.ccy,
            date=now(),
        )

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.is_callable(self.user)


class CurrentMonthRedirectViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``CurrentMonthRedirectView`` view class."""
    view_class = views.CurrentMonthRedirectView

    def setUp(self):
        super(CurrentMonthRedirectViewTestCase, self).setUp()
        self.now_ = now()

    def get_view_kwargs(self):
        return {'year': self.now_.year, 'month': self.now_.month}

    def test_view(self):
        expected_url = reverse(
            'account_keeping_month', kwargs=self.get_view_kwargs())
        self.redirects(expected_url)


class CurrentYearRedirectViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``CurrentYearRedirectView`` view class."""
    view_class = views.CurrentYearRedirectView

    def setUp(self):
        super(CurrentYearRedirectViewTestCase, self).setUp()
        self.now_ = now()

    def get_view_kwargs(self):
        return {'year': self.now_.year, }

    def test_view(self):
        expected_url = reverse(
            'account_keeping_year', kwargs=self.get_view_kwargs())
        self.redirects(expected_url)


class MonthViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``MonthView`` view class."""
    view_class = views.MonthView

    def setUp(self):
        self.user = mixer.blend('auth.User', is_superuser=True)
        self.ccy = mixer.blend('currency_history.Currency', iso_code='EUR')
        self.account = mixer.blend('account_keeping.Account',
                                   currency=self.ccy)
        self.trans1 = mixer.blend(
            'account_keeping.Transaction',
            account=self.account, currency=self.ccy)

        self.ccy2 = mixer.blend('currency_history.Currency')
        self.account2 = mixer.blend('account_keeping.Account',
                                    currency=self.ccy2)
        self.trans2 = mixer.blend(
            'account_keeping.Transaction',
            account=self.account2, currency=self.ccy2)
        rate = mixer.blend(
            'currency_history.CurrencyRateHistory',
            rate__from_currency=self.ccy2,
            rate__to_currency=self.ccy,
        )
        rate.date = now() - timedelta(days=131)
        rate.save()

    def get_view_kwargs(self):
        return {
            'year': self.trans1.transaction_date.year,
            'month': self.trans1.transaction_date.month,
        }

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.is_callable(self.user)
        self.is_callable(self.user, kwargs={
            'year': now().year, 'month': now().month})


class YearOverviewViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``YearOverviewView`` view class."""
    view_class = views.YearOverviewView

    def setUp(self):
        self.user = mixer.blend('auth.User', is_superuser=True)
        self.ccy = mixer.blend('currency_history.Currency', iso_code='EUR')
        self.account = mixer.blend('account_keeping.Account',
                                   currency=self.ccy)
        self.trans1 = mixer.blend(
            'account_keeping.Transaction', transaction_date=now(),
            account=self.account, currency=self.ccy)
        mixer.blend(
            'account_keeping.Transaction', transaction_date=now(),
            account=self.account, currency=self.ccy,
            transaction_type=views.DEPOSIT)

        self.ccy2 = mixer.blend('currency_history.Currency')
        for i in range(11):
            new_date = now().replace(day=1).replace(month=i + 2)
            rate = mixer.blend(
                'currency_history.CurrencyRateHistory',
                rate__from_currency=self.ccy2,
                rate__to_currency=self.ccy,
            )
            rate.date = new_date
            rate.save()
        self.account2 = mixer.blend('account_keeping.Account',
                                    currency=self.ccy2)
        mixer.blend(
            'account_keeping.Transaction', transaction_date=now(),
            transaction_type=views.WITHDRAWAL,
            account=self.account2, currency=self.ccy2)
        mixer.blend(
            'account_keeping.Transaction', transaction_date=now(),
            transaction_type=views.DEPOSIT,
            account=self.account2, currency=self.ccy)

        mixer.blend(
            'account_keeping.Invoice', invoice_date=now(),
            invoice_type=views.DEPOSIT,
            account=self.account2, currency=self.ccy2)

    def get_view_kwargs(self):
        return {'year': now().year, }

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.is_callable(self.user)
        self.is_callable(self.user, kwargs={'year': now().year})


class IndexViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``IndexView`` view class."""
    view_class = views.IndexView

    def setUp(self):
        self.user = mixer.blend('auth.User', is_superuser=True)
        mixer.blend('account_keeping.Transaction')

    @patch('requests.request')
    def test_view(self, mock):
        resp = Response()
        resp.status_code = 200
        resp._content = json.dumps([{
            'reference': 1,
        }])
        mock.return_value = resp
        self.should_redirect_to_login_when_anonymous()
        self.is_callable(self.user)


class PayeeListViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``PayeeListView`` view class."""
    view_class = views.PayeeListView

    def setUp(self):
        self.user = mixer.blend('auth.User', is_superuser=True)
        mixer.blend('account_keeping.Transaction')

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.is_callable(self.user)


class InvoiceCreateViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``InvoiceCreateView`` view class."""
    view_class = views.InvoiceCreateView

    def setUp(self):
        self.user = mixer.blend('auth.User', is_superuser=True)

    def test_view(self):
        self.is_callable(self.user)
        self.is_postable(self.user, data={
            'invoice_type': 'd',
            'invoice_date': now().strftime('%Y-%m-%d'),
            'currency': mixer.blend('currency_history.Currency').pk,
            'amount_net': 0,
            'amount_gross': 0,
            'vat': 0,
            'value_net': 0,
            'value_gross': 0,
        }, to_url_name='account_keeping_month')


class TransactionCreateViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``TransactionCreateView`` view class."""
    view_class = views.TransactionCreateView

    def setUp(self):
        self.user = mixer.blend('auth.User', is_superuser=True)

    def test_view(self):
        self.is_callable(self.user)
        date = now()
        account = mixer.blend('account_keeping.Account')
        data = {
            'transaction_type': 'd',
            'transaction_date': date.strftime('%Y-%m-%d'),
            'account': account.pk,
            'payee': mixer.blend('account_keeping.Payee').pk,
            'category': mixer.blend('account_keeping.Category').pk,
            'currency': mixer.blend('currency_history.Currency').pk,
            'amount_net': 0,
            'amount_gross': 0,
            'vat': 0,
            'value_net': 0,
            'value_gross': 0,
        }
        self.is_postable(self.user, data=data, to=u'{}#{}'.format(
            reverse('account_keeping_month', kwargs={
                'year': date.year, 'month': date.month}), account.slug))

        self.is_callable(self.user, data={'invoice': 1})
        self.is_callable(self.user, data={'parent': 1})


class TransactionExportViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``TransactionExportView`` view class."""
    view_class = views.TransactionExportView

    def setUp(self):
        self.user = mixer.blend('auth.User', is_superuser=True)

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.is_callable(self.user)
        self.is_postable(user=self.user, ajax=True, data={
            'account': mixer.blend('account_keeping.Account').pk,
            'start': '2015-01-01',
            'end': '2018-01-01',
        })



"""Forms of the account_keeping app."""
from django import forms
from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _

from . import models


class InvoiceForm(forms.ModelForm):
    class Meta:
        model = models.Invoice
        fields = '__all__'
        try:
            widgets = {
                'invoice_date': forms.widgets.SelectDateWidget(
                    attrs={'style': 'display: inline; width: auto;'}),
                'payment_date': forms.widgets.SelectDateWidget(
                    attrs={'style': 'display: inline; width: auto;'}),
            }
        except AttributeError:  # pragma: nocover
            widgets = {
                'invoice_date': forms.widgets.DateInput,
                'payment_date': forms.widgets.DateInput,
            }


class TransactionForm(forms.ModelForm):
    mark_invoice = forms.BooleanField(
        label=_('Mark invoice as paid?'),
        initial=True,
        required=False,
        widget=forms.widgets.CheckboxInput(
            attrs={'data-id': 'mark-invoice-field'}),
    )

    class Meta:
        model = models.Transaction
        fields = '__all__'
        try:
            date_widget = forms.widgets.SelectDateWidget(
                attrs={'style': 'display: inline; width: auto;'})
        except AttributeError:  # pragma: nocover
            date_widget = forms.widgets.DateInput
        widgets = {
            'transaction_date': date_widget,
            'invoice': forms.widgets.NumberInput(
                attrs={'data-id': 'invoice-field'}),
            'parent': forms.widgets.NumberInput(),
        }

    def __init__(self, *args, **kwargs):
        super(TransactionForm, self).__init__(*args, **kwargs)
        self.fields['payee'].help_text = _(
            '<a href="{}">Add a payee</a>').format(
            reverse('account_keeping_payee_create'))

    def save(self, *args, **kwargs):
        if self.instance.invoice and self.cleaned_data.get('mark_invoice'):
            # Set the payment date on related invoice
            self.instance.invoice.payment_date = self.instance.transaction_date
            self.instance.invoice.save()
        return super(TransactionForm, self).save(*args, **kwargs)


class ExportForm(forms.Form):
    account = forms.ModelChoiceField(
        label=_('Account'),
        queryset=models.Account.objects.all(),
    )

    start = forms.DateField(
        label=_('Start'),
    )

    end = forms.DateField(
        label=_('End'),
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-05-30 08:01
from __future__ import unicode_literals

import account_keeping.models
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('currency_history', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128)),
                ('slug', models.SlugField(max_length=128)),
                ('initial_amount', models.DecimalField(decimal_places=2, default=0, max_digits=10)),
                ('total_amount', models.DecimalField(decimal_places=2, default=0, max_digits=10)),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='accounts', to='currency_history.Currency')),
            ],
        ),
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Invoice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('invoice_type', models.CharField(choices=[(b'w', b'withdrawal'), (b'd', b'deposit')], max_length=1)),
                ('invoice_date', models.DateField()),
                ('invoice_number', models.CharField(blank=True, max_length=256)),
                ('description', models.TextField(blank=True)),
                ('amount_net', models.DecimalField(blank=True, decimal_places=2, default=0, max_digits=10)),
                ('vat', models.DecimalField(decimal_places=2, default=0, max_digits=4)),
                ('amount_gross', models.DecimalField(blank=True, decimal_places=2, default=0, max_digits=10)),
                ('value_net', models.DecimalField(decimal_places=2, default=0, max_digits=10)),
                ('value_gross', models.DecimalField(decimal_places=2, default=0, max_digits=10)),
                ('payment_date', models.DateField(blank=True, null=True)),
                ('pdf', models.FileField(blank=True, null=True, upload_to=b'invoice_files')),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='invoices', to='currency_history.Currency')),
            ],
            options={
                'ordering': ['-invoice_date'],
            },
            bases=(account_keeping.models.AmountMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Payee',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('transaction_type', models.CharField(choices=[(b'w', b'withdrawal'), (b'd', b'deposit')], max_length=1)),
                ('transaction_date', models.DateField()),
                ('description', models.TextField(blank=True)),
                ('invoice_number', models.CharField(blank=True, max_length=256)),
                ('amount_net', models.DecimalField(blank=True, decimal_places=2, default=0, max_digits=10)),
                ('vat', models.DecimalField(decimal_places=2, default=0, max_digits=4)),
                ('amount_gross', models.DecimalField(blank=True, decimal_places=2, default=0, max_digits=10)),
                ('value_net', models.DecimalField(decimal_places=2, default=0, max_digits=10)),
                ('value_gross', models.DecimalField(decimal_places=2, default=0, max_digits=10)),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transactions', to='account_keeping.Account')),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transactions', to='account_keeping.Category')),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transactions', to='currency_history.Currency')),
                ('invoice', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='transactions', to='account_keeping.Invoice')),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='account_keeping.Transaction')),
                ('payee', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='transactions', to='account_keeping.Payee')),
            ],
            options={
                'ordering': ['-transaction_date'],
            },
            bases=(account_keeping.models.AmountMixin, models.Model),
        ),
    ]



"""URLs to run the tests."""
from django.conf.urls import include, url
from django.contrib import admin


admin.autodiscover()

urlpatterns = [
    url(r'^admin/', include(admin.site.urls)),
    url(r'^', include('account_keeping.urls')),
]



"""URLs for the account_keeping app."""
from django.conf.urls import url

from . import views


urlpatterns = [
    url(r'transaction/(?P<pk>\d+)/$',
        views.TransactionUpdateView.as_view(),
        name='account_keeping_transaction_update'),

    url(r'transaction/create/$',
        views.TransactionCreateView.as_view(),
        name='account_keeping_transaction_create'),

    url(r'invoice/(?P<pk>\d+)/$',
        views.InvoiceUpdateView.as_view(),
        name='account_keeping_invoice_update'),

    url(r'invoice/create/$',
        views.InvoiceCreateView.as_view(),
        name='account_keeping_invoice_create'),

    url(r'accounts/$',
        views.AccountListView.as_view(),
        name='account_keeping_accounts'),

    url(r'payees/(?P<pk>\d+)/$',
        views.PayeeUpdateView.as_view(),
        name='account_keeping_payee_update'),

    url(r'payees/create/$',
        views.PayeeCreateView.as_view(),
        name='account_keeping_payee_create'),

    url(r'payees/$',
        views.PayeeListView.as_view(),
        name='account_keeping_payees'),

    url(r'export/$',
        views.TransactionExportView.as_view(),
        name='account_keeping_export'),

    url(r'all/$',
        views.AllTimeView.as_view(),
        name='account_keeping_all'),

    url(r'(?P<year>\d+)/(?P<month>\d+)/$',
        views.MonthView.as_view(),
        name='account_keeping_month'),

    url(r'(?P<year>\d+)/$',
        views.YearOverviewView.as_view(),
        name='account_keeping_year'),

    url(r'current-year/$',
        views.CurrentYearRedirectView.as_view(),
        name='account_keeping_current_year'),

    url(r'current-month/$',
        views.CurrentMonthRedirectView.as_view(),
        name='account_keeping_current_month'),

    url(r'$',
        views.IndexView.as_view(),
        name='account_keeping_index'),
]



"""Tests for the utils of the account_keeping app."""
import datetime

from django.test import TestCase

from .. import utils


class GetDateTestCase(TestCase):
    """Tests for the ``get_date`` function."""
    longMessage = True

    def test_function(self):
        self.assertEqual(utils.get_date('2014-01-01'),
                         datetime.datetime(2014, 1, 1))
        self.assertEqual(utils.get_date(1), 1)


class GetMonthsOfYearTestCase(TestCase):
    """Tests for the ``get_months_of_year`` function."""
    longMessage = True

    def test_function(self):
        self.assertEqual(
            utils.get_months_of_year(datetime.datetime.now().year - 1), 12)
        self.assertEqual(
            utils.get_months_of_year(datetime.datetime.now().year + 1), 1)
        self.assertEqual(
            utils.get_months_of_year(datetime.datetime.now().year),
            datetime.datetime.now().month)



"""
Importer that handles .csv files exported via Money Manager Ex.

IMPORTANT: MMEX does not distinguish between incoming and outgoing "Transfer"
transactions. After you export the .csv you must identify all incoming
"Transfer" transactions and rename the type to "TransferDeposit".

"""
from decimal import Decimal
from optparse import make_option
import csv
import datetime

from django.core.management.base import BaseCommand, CommandError

from currency_history.models import Currency

from ... import models


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option(
            '-f', '--file',
            dest='filepath',
            help='Filepath to the .csv file that contains the data'),
        make_option(
            '-a', '--account',
            dest='account',
            help='Account slug of the account that should hold the new data'),
        make_option(
            '-c', '--currency',
            dest='currency',
            help='ISO-code of the currency for the specified data'),
        make_option(
            '-t', '--vat',
            dest='vat',
            help='VAT that should be applied to all transactions (i.e. 19)'),
    )
    help = 'Imports the specified .csv file from mmex'

    def handle(self, *args, **options):
        try:
            currency = Currency.objects.get(iso_code=options.get('currency'))
        except Currency.DoesNotExist:
            raise CommandError('The specified currency does not exist')

        account = models.Account.objects.get(slug=options.get('account'))
        vat = options.get('vat')
        if not vat:
            vat = 0
        vat = Decimal(vat)

        filepath = options.get('filepath')

        deposit = models.Transaction.TRANSACTION_TYPES['deposit']
        withdrawal = models.Transaction.TRANSACTION_TYPES['withdrawal']

        with open(filepath, 'rt') as csvfile:
            csvreader = csv.reader(csvfile)
            for row in csvreader:
                transaction_date = datetime.datetime.strptime(
                    row[0], '%d/%m/%Y')
                payee, created = models.Payee.objects.get_or_create(
                    name=row[1])
                if row[2] in ['Withdrawal', 'Transfer']:
                    transaction_type = withdrawal
                else:
                    transaction_type = deposit
                amount = Decimal(row[3])
                if row[4] and not row[5]:
                    cat_name = row[4]
                else:
                    cat_name = row[5]
                category, created = models.Category.objects.get_or_create(
                    name=cat_name)
                description = row[7]

                invoice = models.Invoice.objects.create(
                    invoice_type=transaction_type,
                    invoice_date=datetime.date(1900, 1, 1),
                    currency=currency,
                    amount_gross=amount,
                    vat=vat,
                    payment_date=transaction_date,
                )

                models.Transaction.objects.create(
                    account=account,
                    transaction_type=transaction_type,
                    transaction_date=transaction_date,
                    description=description,
                    invoice=invoice,
                    payee=payee,
                    category=category,
                    currency=currency,
                    amount_gross=amount,
                    vat=vat,
                )



"""Admin classes for the account_keeping app."""
from django.contrib import admin

from . import models


class AccountAdmin(admin.ModelAdmin):
    list_display = [
        'name', 'slug', 'currency', 'initial_amount', 'total_amount']
admin.site.register(models.Account, AccountAdmin)


class InvoiceAdmin(admin.ModelAdmin):
    list_display = [
        'invoice_type', 'invoice_date', 'currency', 'amount_net', 'vat',
        'amount_gross', 'payment_date']
    list_filter = ['invoice_type', 'currency', 'payment_date']
    date_hierarchy = 'invoice_date'
    search_fields = ['invoice_number', 'description']
admin.site.register(models.Invoice, InvoiceAdmin)


class PayeeAdmin(admin.ModelAdmin):
    list_display = ['name', ]
admin.site.register(models.Payee, PayeeAdmin)


class CategoryAdmin(admin.ModelAdmin):
    list_display = ['name', ]
admin.site.register(models.Category, CategoryAdmin)


class TransactionAdmin(admin.ModelAdmin):
    list_display = [
        'transaction_date', 'parent', 'invoice_number', 'invoice', 'payee',
        'category', 'currency', 'value_net', 'vat', 'value_gross', ]
    list_filter = ['account', 'currency', 'payee', 'category']
    date_hierarchy = 'transaction_date'
    raw_id_fields = ['parent', 'invoice']
    search_fields = [
        'invoice_number', 'invoice__invoice_number', 'description']
admin.site.register(models.Transaction, TransactionAdmin)



"""Tests for the freckle API functions of the account_keeping app."""
import json

from django.test import TestCase

from mixer.backend.django import mixer
from mock import patch
from requests import Response

from .. import freckle_api


class GetUnpaidInvoicesWithTransactionsTestCase(TestCase):
    """Tests for the ``get_unpaid_invoices_with_transactions`` function."""
    longMessage = True

    @patch('requests.request')
    def test_function(self, mock):
        invoice = mixer.blend('account_keeping.Invoice')
        mixer.blend('account_keeping.Transaction', invoice=invoice)
        resp = Response()
        resp.status_code = 200
        resp._content = json.dumps([{
            'reference': invoice.invoice_number,
        }])
        mock.return_value = resp
        self.assertEqual(
            len(freckle_api.get_unpaid_invoices_with_transactions()), 1)



"""Tests for the template tags of the account_keeping app."""
from django.test import TestCase

from ..templatetags import account_keeping_tags


class CurrencyTestCase(TestCase):
    """Tests for the ``currency`` filter."""
    longMessage = True

    def test_tag(self):
        self.assertEqual(account_keeping_tags.currency('1.11111'), 'EUR 1.11')



"""
Collects invoices for the given account and timeframe into a tarball.

"""
from optparse import make_option
import datetime
import os
import shutil

from django.core.management.base import BaseCommand

from ... import models


class Command(BaseCommand):
    option_list = BaseCommand.option_list + (
        make_option(
            '-o', '--output',
            dest='output',
            help='Output folder. Make sure that the folder exists.'),
        make_option(
            '-a', '--account',
            dest='account',
            help='Account slug of the account that should be handled.'),
        make_option(
            '-s', '--start',
            dest='start_date',
            help='Start date. Include all transactions from this date.'),
        make_option(
            '-e', '--end',
            dest='end_date',
            help='End date. Include all transactions up to this date.'),
    )
    help = 'Copies invoices of transactions into a folder.'

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.counter = 1

    def copy_file(self, transaction):
        if transaction.invoice and transaction.invoice.pdf:  # pragma: nocover
            counter = str(self.counter).zfill(4)
            filename = '{0}.pdf'.format(counter)
            shutil.copy(
                transaction.invoice.pdf.file.name,
                os.path.join(self.output_folder, filename))
            self.counter += 1

    def handle(self, *args, **options):
        account = models.Account.objects.get(slug=options.get('account'))
        self.output_folder = options.get('output')
        start_date = datetime.datetime.strptime(
            options.get('start_date'), '%Y-%m-%d')
        end_date = datetime.datetime.strptime(
            options.get('end_date'), '%Y-%m-%d')
        transactions = models.Transaction.objects.filter(
            account=account,
            transaction_date__gte=start_date,
            transaction_date__lte=end_date,
        ).prefetch_related('children', ).order_by('-transaction_date')
        for transaction in transactions:
            if transaction.children.all():
                for child in transaction.children.all():
                    self.copy_file(child)
            else:
                self.copy_file(transaction)
